# -*- coding: utf-8 -*-
from functools import wraps
import logging

from sanic import Sanic

from .exceptions import error_handler
from .filters import filter_validate_schemas, filter_validate_response
from .http import Request

log = logging.getLogger(__name__)

__all__ = ('Rafter',)


class Rafter(Sanic):
    """
    This class inherits Sanic's default ``Sanic`` class.
    It provides an instance of your app, to which you can add
    resources or regular Sanic routes.
    """

    default_filters = [
        filter_validate_schemas,
        filter_validate_response,
    ]
    """
    Default filters called on every resource route.
    """

    default_error_handlers = ((Exception, error_handler),)
    """
    Default error handlers. It must be a list of tuples containing the
    exception type and a callable.
    """

    default_request_class = Request
    """
    The default request class. If changed, it must inherit from
    :class:`rafter.http.Request`.
    """

    def __init__(self, **kwargs):
        """
        .. note::
            Please refer to `Sanic API reference <sanic_ref_>`_
            for init arguments.
        """
        kwargs.setdefault('request_class', self.default_request_class)
        if not issubclass(kwargs['request_class'], Request):
            raise RuntimeError('request_class should inherit '
                               'rafter.http.Request class')

        super(Rafter, self).__init__(**kwargs)

        for e, f in self.default_error_handlers:
            self.error_handler.add(e, f)

    def resource(self, uri, methods=frozenset({'GET'}), **kwargs):
        """
        Decorates a function to be registered as a resource route.

        :param uri: path of the URL
        :param methods: list or tuple of methods allowed
        :param filters: List of callable that will filter request and
                        response data
        :param validators: List of callable added to the filter list.
        :param request_schema: Schema for request data
        :param response_schema: Schema for response data
        :param host:
        :param strict_slashes:
        :param stream:
        :param version:
        :param name: user defined route name for url_for

        :return: A decorated function
        """
        def decorator(f):
            self.add_resource(f, uri, methods, **kwargs)

        return decorator

    def add_resource(self, handler, uri, methods=frozenset({'GET'}),
                     filters: [callable]=default_filters,
                     validators: [callable]=[],
                     request_schema=None,
                     response_schema=None,
                     **kwargs):
        """
        Register a resource route.

        :param handler: function or class instance
        :param uri: path of the URL
        :param methods: list or tuple of methods allowed
        :param filters: List of callable that will filter request and
                        response data
        :param validators: List of callable added to the filter list.
        :param request_schema: Schema for request data
        :param response_schema: Schema for response data
        :param host:
        :param strict_slashes:
        :param stream:
        :param version:
        :param name: user defined route name for url_for

        :return: function or class instance
        """

        filter_list = filters + validators
        filter_options = {
            'request_schema': request_schema,
            'response_schema': response_schema
        }
        filter_options.update(kwargs)

        handler = self.init_filters(filter_list, filter_options)(handler)
        return self.add_route(handler, uri, methods, **kwargs)

    @staticmethod
    def init_filters(filter_list, params):
        def decorator(handler):
            get_response = handler
            for f in filter_list:
                get_response = f(get_response, params)

            @wraps(handler)
            async def wrapper(request, *args, **kwargs):
                return await get_response(request, *args, **kwargs)

            return wrapper
        return decorator



# -*- coding: utf-8 -*-
from sanic.request import Request as BaseRequest
from sanic.response import HTTPResponse, json_dumps

__all__ = ('Response',)


class Request(BaseRequest):
    """
    This class is the default :class:`rafter.app.Rafter`'s request object
    that will be transmitted to every route. It adds a ``validated``
    attribute that will contains all of the validated values, if the route
    uses schemas.

    .. py:attribute:: validated

            This property can contain the request data after validation
            and conversion by the filter.
    """

    def __init__(self, *args, **kwargs):
        super(Request, self).__init__(*args, **kwargs)

        self.validated = {}


class Response(HTTPResponse):
    """
    A response object that you can return in any route. It looks a lot like
    ``sanic.response.json`` function. With a few differences:

    - Its data will be serialized at the very last moment
    - Its data might be validated with
      :func:`rafter.filters.filter_validate_schemas`.

    Example:

    .. code-block:: python
        :emphasize-lines: 3

        @app.resource('/')
        def main_route(request):
            return Response({'data': 'some data'})
    """

    def __init__(self, body=None, status=200, headers=None,
                 content_type='application/json'):
        """
        :param body: The data this will be serialized in response.
        :param status: The response status code.
        :param headers: Additionnal headers.
        :param content_type: Response MIME Type

        .. important::
            Input ``body`` will later be serialized. Its value is held by
            the internal ``_data`` attribute.
            Settings the ``body`` value will set this internal data but
            getting ``body`` will return the serialized value. This is the
            only way we can force Sanic HTTPResponse to serialize our response
            as late as possible.
        """
        super(Response, self).__init__(body=None, status=status,
                                       headers=headers,
                                       content_type=content_type)
        self._data = body

    @property
    def body(self):
        return self._encode_body(json_dumps(self._data))

    @body.setter
    def body(self, data):
        self._data = data

    @property
    def data(self):
        return self._data



# -*- coding: utf-8 -*-
from copy import deepcopy
import logging

from sanic.exceptions import abort
from sanic.request import RequestParameters
from sanic.response import HTTPResponse
from sanic.server import CIDict
from schematics.exceptions import BaseError
from schematics.types import ListType

from .exceptions import ValidationErrors
from .http import Response

log = logging.getLogger(__name__)


def filter_validate_schemas(get_response, params):
    """
    This filter validates input data against the resource's
    ``request_schema`` and fill the request's ``validated`` dict.

    Data from ``request.params`` and ``request.body`` (when the request body
    is of a form type) will be converted using the schema in order to get
    proper lists or unique values.

    .. important::
        The request validation is only effective when a
        ``request_schema`` has been provided by the resource definition.
    """

    request_schema = params.get('request_schema')

    if request_schema is None:
        return get_response

    def _convert_params(schema, data):
        for sc in schema.fields.values():
            name = sc.serialized_name or sc.name
            val = data.getlist(name)
            if val is None:
                continue

            if len(val) == 1 and not isinstance(sc, ListType):
                val = val[0]

            data[name] = val

    async def decorated_filter(request, *args, **kwargs):
        data = {
            'headers': CIDict(request.headers),
            'path': request.app.router.get(request)[2],
            'params': RequestParameters(request.args),
            'body': {}
        }

        if request.body:
            # Get body if we have something there
            if request.form:
                data['body'] = RequestParameters(request.form)
            else:
                # will raise 400 if cannot parse json
                data['body'] = deepcopy(request.json)

        if hasattr(request_schema, 'body') and request.form:
            _convert_params(request_schema.body, data['body'])

        if hasattr(request_schema, 'params') and data['params']:
            _convert_params(request_schema.params, data['params'])

        # Now, validate the whole thing
        try:
            model = request_schema(data, strict=False, validate=False)
            model.validate()
            request.validated = model.to_native()
        except BaseError as e:
            raise ValidationErrors(e.to_primitive())

        return await get_response(request, *args, **kwargs)

    return decorated_filter


def filter_validate_response(get_response, params):
    """
    This filter process the returned response. It does 3 things:

    - If the response is a ``sanic.response.HTTPResponse`` and not a
      :class:`rafter.http.Response`, return it immediately.
    - If the response is not a :class:`rafter.http.Response` instance,
      turn it to a :class:`rafter.http.Response` instance.
    - Then, process, validate and serialize this response when necessary.

    That means that you can always return a normal Sanic's HTTPResponse
    and thus, bypass the validation process when you need to do so.

    .. important::
        The response validation is only effective when:

        - A ``response_schema`` has been provided by the resource definition
        - The resource returns a :class:`rafter.http.Response` instance
          or arbitrary data.
    """

    schema = params.get('response_schema')

    async def decorated_filter(request, *args, **kwargs):
        response = await get_response(request, *args, **kwargs)

        if isinstance(response, HTTPResponse) and \
                not isinstance(response, Response):
            return response

        if not isinstance(response, Response):
            response = Response(response)

        if schema:
            try:
                model = schema(response.data, strict=False, validate=False)
                model.validate()
                response.body = model.to_primitive()
            except BaseError as e:
                log.exception(e)
                abort(500, 'Wrong data output')

        return response

    return decorated_filter



# -*- coding: utf-8 -*-
import logging
from operator import itemgetter
import sys
import traceback

from sanic.exceptions import SanicException
from sanic.response import json

__all__ = ('ApiError', 'ValidationErrors')

log = logging.getLogger(__name__)


def error_handler(request, exception):
    """
    A generic exception handler. Returns a JSON response
    with structured error data.
    """
    data = {'status': 500,
            'message': 'An error occured'}

    if isinstance(exception, SanicException):
        data.update({'status': getattr(exception, 'status_code', 500),
                     'message': str(exception)})

    if isinstance(exception, ApiError):
        data.update(exception.to_primitive())

    if data['status'] >= 500:
        log.exception(exception)
        if request.app.debug:  # pragma: nocover
            exc_type, exc_value, exc_traceback = sys.exc_info()
            data['stack'] = traceback.extract_tb(exc_traceback)

    return json(data, status=data['status'])


class ApiError(SanicException):
    def __init__(self, message: str, status_code: int=500, **kwargs):
        """
        :param message: An error message
        :param status_code: The returned HTTP status code

        Any other keyword argument will be added to the error data and
        will be added to the return of
        :attr:`~rafter.exceptions.ApiError.data`.
        """
        super(ApiError, self).__init__(message, status_code=status_code)
        self._data = kwargs

    @property
    def data(self):
        """
        Returns the internal property ``_data``. It can be overriden by
        specialized inherited exceptions.
        """
        return self._data

    def to_primitive(self) -> dict:
        """
        This methods is called by the default error handler
        :func:`error_handler` and returns a dict of error data.
        """
        return self.data or {}


class ValidationErrors(ApiError):
    def __init__(self, errors: dict, **kwargs):
        super(ValidationErrors, self).__init__('Invalid input data', 400,
                                               **kwargs)
        self._errors = errors

    @property
    def data(self):
        """
        Returns a dictionnary containing all the passed data and an item
        ``error_list`` which holds the result of :attr:`error_list`.
        """
        res = {'error_list': self.error_list}
        res.update(super(ValidationErrors, self).data)
        return res

    @property
    def error_list(self):
        """
        Returns an error list based on the internal error dict values.
        Each item contains a dict with ``messages`` and ``path`` keys.

        Example::

            >>> errors = {
            >>>     'body': {
            >>>         'age': ['invalid age'],
            >>>         'options': {
            >>>             'extra': {
            >>>                 'ex1': ['invalid ex1'],
            >>>             }
            >>>         }
            >>>     }
            >>> }
            >>> e = ValidationErrors(errors)
            >>> e.error_list

            [
                {'messages': ['invalid age'],
                 'location': ['body', 'age']},
                {'messages': ['invalid ex1'],
                 'location': ['body', 'options', 'extra', 'ex1']},
            ]
        """
        return sorted(self._walk(self._errors), key=itemgetter('location'))

    def _walk(self, d, path=[]):
        nested_keys = tuple(k for k in d.keys() if isinstance(d[k], dict))
        items = tuple((k, d[k]) for k in d.keys() if k not in nested_keys)

        for p, v in items:
            yield {
                'location': path + [p],
                'messages': v
            }

        for k in nested_keys:
            for res in self._walk(d[k], path + [k]):
                yield res



# -*- coding: utf-8 -*-
import os.path
import re
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

# Get app version
with open(os.path.join(here, 'rafter/__init__.py'), 'r') as fp:
    version = re.compile(r".*__version__ = '(.*?)'",
                         re.S).match(fp.read()).group(1)


# Read README file
def readme():
    with open('README.md', 'r') as fp:
        return fp.read()


packages = find_packages(exclude=['tests'])

requirements = (
    'sanic >= 0.7.0',
    'schematics',
    'ujson',
)

extras_require = {
    'setup': (
        'pytest-runner',
    ),
    'test': (
        'pytest',
        'pytest-cov',
        'pytest-sanic',
    )
}

# Setup
setup(
    name='rafter',
    version=version,
    description='Building blocks for REST APIs on top of Sanic',
    author='Olivier Meunier',
    author_email='olivier@neokraft.net',
    license='MIT',
    url='https://github.com/olivier-m/rafter',
    project_urls={
        'Documentation': 'https://rafter.readthedocs.io/',
        'Source': 'https://github.com/olivier-m/rafter',
        'Tracker': 'https://github.com/olivier-m/rafter/issues',
    },
    long_description=readme(),
    keywords='rest sanic framework',
    install_requires=requirements,
    extras_require=extras_require,
    setup_requires=extras_require['setup'],
    tests_require=extras_require['test'],
    packages=packages,
    zip_safe=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers'
    ],
)



# -*- coding: utf-8 -*-
from schematics import types

__all__ = ('model_node',)


def model_node(**kwargs):
    """
    Decorates a ``schematics.Model`` class to add it as a field
    of type ``schematic.types.ModelType``.

    Keyword arguments are passed to ``schematic.types.ModelType``.

    Example:

    .. code-block:: python
        :emphasize-lines: 6,11

        from schematics import Model, types

        class MyModel(Model):
            name = types.StringType()

            @model_node()
            class options(Model):
                status = types.IntType()

            # With arguments and another name
            @model_node(serialized_name='extra', required=True)
            class _extra(Model):
                test = types.StringType()
    """
    kwargs.setdefault('default', {})

    def decorator(model):
        return types.ModelType(model, **kwargs)

    return decorator
